/***************************************************************************
                          collision.c  -  description
                             -------------------
    begin                : Sun Apr 6 2003
    copyright            : (C) 2003 by Steven Goodwin
    email                : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <malloc.h>
#include "explore.h"
#include "level.h"
#include "objects.h"
#include "collision.h"

/* Faster Line Segment Intersection   */
/* Franklin Antonio                   */

/* return values */
#define DONT_INTERSECT 0
#define DO_INTERSECT   1
#define COLLINEAR      2


/* The SAME_SIGNS macro assumes arithmetic where the exclusive-or    */
/* operation will work on sign bits.  This works for twos-complement,*/
/* and most other machine arithmetic.                                */
#define SAME_SIGNS( a, b ) \
	(((long) ((unsigned long) a ^ (unsigned long) b)) >= 0 )


/* The use of some short working variables allows this code to run   */
/* faster on 16-bit computers, but is not essential.  It should not  */
/* affect operation on 32-bit computers.  The short working variables*/
/* to not restrict the range of valid input values, as these were    */
/* constrained in any case, due to algorithm restrictions.           */

static int lines_intersect(int x1,int y1,int x2,int y2,int x3,int y3,int x4,int y4,int *x,int *y)
{

long Ax,Bx,Cx,Ay,By,Cy,d,e,f,num,offset;
short x1lo,x1hi,y1lo,y1hi;

Ax = x2-x1;
Bx = x3-x4;

if(Ax<0) {						/* X bound box test*/
  x1lo=(short)x2; x1hi=(short)x1;
  } else {
  x1hi=(short)x2; x1lo=(short)x1;
  }
if(Bx>0) {
  if(x1hi < (short)x4 || (short)x3 < x1lo) return DONT_INTERSECT;
  } else {
  if(x1hi < (short)x3 || (short)x4 < x1lo) return DONT_INTERSECT;
  }

Ay = y2-y1;
By = y3-y4;

if(Ay<0) {						/* Y bound box test*/
  y1lo=(short)y2; y1hi=(short)y1;
  } else {
  y1hi=(short)y2; y1lo=(short)y1;
  }
if(By>0) {
  if(y1hi < (short)y4 || (short)y3 < y1lo) return DONT_INTERSECT;
  } else {
  if(y1hi < (short)y3 || (short)y4 < y1lo) return DONT_INTERSECT;
  }


Cx = x1-x3;
Cy = y1-y3;
d = By*Cx - Bx*Cy;					/* alpha numerator*/
f = Ay*Bx - Ax*By;					/* both denominator*/
if(f>0) {						/* alpha tests*/
  if(d<0 || d>f) return DONT_INTERSECT;
  } else {
  if(d>0 || d<f) return DONT_INTERSECT;
  }

e = Ax*Cy - Ay*Cx;					/* beta numerator*/
if(f>0) {						/* beta tests*/
  if(e<0 || e>f) return DONT_INTERSECT;
  } else {
  if(e>0 || e<f) return DONT_INTERSECT;
  }

/*compute intersection coordinates*/

if(f==0) return COLLINEAR;
num = d*Ax;						/* numerator */
offset = SAME_SIGNS(num,f) ? f/2 : -f/2;		/* round direction*/
*x = x1 + (num+offset) / f;				/* intersection x */

num = d*Ay;
offset = SAME_SIGNS(num,f) ? f/2 : -f/2;
*y = y1 + (num+offset) / f;				/* intersection y */

return DO_INTERSECT;
}
/*-----------------------------------------------------------------*/
/* -- end of copied section -- The above code is not actually used */
/* in Explorer Dug, but left in to show the sort of collision code */
/* that may appear in game code from time to time!!                */

BOOL exDoBoxesIntersect(const SDL_Rect *p1, const SDL_Rect *p2)
{
	if (p1->x >= p2->x+p2->w)	return FALSE;
	if (p1->x+p1->w <= p2->x)	return FALSE;
	if (p1->y >= p2->y+p2->h)	return FALSE;
	if (p1->y+p1->h <= p2->y)	return FALSE;

return TRUE;
}

void  exBuildRect4Tile(SDL_Rect *pRc, int tx, int ty)
{
EX_TILE *pTile;

	pRc->x = tx*TheGame.iTileWidth;
	pRc->y = ty*TheGame.iTileWidth;

	pTile = exGetTileAt(tx, ty);
	pRc->w = pTile->iCollideWidth;
	pRc->h = pTile->iCollideHeight;
}

void exInitCollisionList(EX_COLLISION_LIST *pCList)
{
int i;

	pCList->iSize = 0;
	pCList->bCheckTiles = TRUE;
	for(i=0;i<eLastObj;i++)
		pCList->bCheckObjects[i] = TRUE;

	pCList->iMax = sizeof(pCList->List)/sizeof(pCList->List[0]);
}

void exAddCollision(EX_COLLISION_LIST *pCList, const EX_COLLISION *pCurr)
{
	if (pCList == NULL || pCList->iSize == pCList->iMax)
		return;

	pCList->List[pCList->iSize] = *pCurr;
	++pCList->iSize;
}
		



BOOL exCCBoundingBox(const SDL_Rect *pRc, const SDL_Rect *pEnemy, const EX_OBJECT *pObj, EX_COLLISION_LIST *pCList)
{
BOOL rt = FALSE;

	if (exDoBoxesIntersect(pRc, pEnemy))
		{
/* find the intersection points */
SDL_Rect r1, r2;
EX_COLLISION col;
int i,j;

 #define DONT_INTERSECT 0
#define DO_INTERSECT   1
#define COLLINEAR      2

/*
+---0---+
|       |
3       1
|       |
+---2---+
*/
for(j=0;j<4;j++)	/* r1 */
{
switch(j) {
	case 0:	r1.x=pRc->x; r1.y=pRc->y; r1.w=r1.x+pRc->w-1; r1.h=r1.y; break;
	case 1:	r1.x=pRc->x+pRc->w-1; r1.y=pRc->y; r1.w=r1.x; r1.h=r1.y+pRc->h-1; break;
	case 2:	r1.x=pRc->x; r1.y=pRc->y+pRc->h-1; r1.w=r1.x+pRc->w-1; r1.h=r1.y; break;
	case 3:	r1.x=pRc->x; r1.y=pRc->y; r1.w=r1.x; r1.h=r1.y+pRc->h-1; break;
}

for(i=0;i<4;i++)
{
switch(i) {
	case 0:	r2.x=pEnemy->x; r2.y=pEnemy->y; r2.w=r2.x+pEnemy->w-1; r2.h=r2.y; break;
	case 1:	r2.x=pEnemy->x+pEnemy->w-1; r2.y=pEnemy->y; r2.w=r2.x; r2.h=r2.y+pEnemy->h-1; break;
	case 2:	r2.x=pEnemy->x; r2.y=pEnemy->y+pEnemy->h-1; r2.w=r2.x+pEnemy->w-1; r2.h=r2.y; break;
	case 3:	r2.x=pEnemy->x; r2.y=pEnemy->y; r2.w=r2.x; r2.h=r2.y+pEnemy->h-1; break;
}

if (lines_intersect(r1.x,r1.y,r1.w,r1.h,  r2.x,r2.y,r2.w,r2.h, &col.dx, &col.dy))
	{
	col.pObj = (EX_OBJECT *)pObj;
	col.edge = j;	/* use players edge */
	exAddCollision(pCList, &col);
	rt = TRUE;
	}
}

}
		}

return rt;
}

int exGetLineCollision(const EX_OBJECT *pObj, int from_x, int from_y, int to_x, int to_y, EX_COLLISION_LIST *pCList)
{
/* Might find a use for this later - but until we do, there's no point in writing the code! */
return 0;
}

BOOL TestTileCollision(const SDL_Rect *pRc, int tx, int ty, EX_COLLISION_LIST *pCList)
{
SDL_Rect nme;
EX_TILE *pTile;

	/* Is tile collidable? */
	pTile = exGetTileAt(tx, ty);
	if (pTile==NULL || pTile->iRegion == 0)
		return FALSE;
	
      /* Get area */
	exBuildRect4Tile(&nme, tx, ty);

	return exCCBoundingBox(pRc, &nme, NULL, pCList);
}

int exGetRectCollision(const EX_OBJECT *pIgnoreObj, const SDL_Rect *pRc, EX_COLLISION_LIST *pCollisionList)
{
	pCollisionList->pIgnoreObj = (EX_OBJECT *)pIgnoreObj;
	pCollisionList->iType = eBoundingBox;
	pCollisionList->rc = *pRc;
	return exCheckCollision(pCollisionList);
}

int exCheckCollision(EX_COLLISION_LIST *pCollisionList)
{
EX_OBJECT *pObj;
SDL_Rect nme;
int tx, ty;
const SDL_Rect *pCollisionRc = &pCollisionList->rc;

	/* Check static world */
	if (pCollisionList->bCheckTiles)
		{
		tx = pCollisionRc->x/TheGame.iTileWidth;
		ty = pCollisionRc->y/TheGame.iTileHeight;

		TestTileCollision(pCollisionRc, tx, ty, pCollisionList);
		if (pCollisionRc->x%TheGame.iTileWidth)
			TestTileCollision(pCollisionRc, tx+1, ty, pCollisionList);
		if (pCollisionRc->y%TheGame.iTileHeight)
			TestTileCollision(pCollisionRc, tx, ty+1, pCollisionList);
		if ((pCollisionRc->x%TheGame.iTileWidth) && (pCollisionRc->y%TheGame.iTileHeight))
			TestTileCollision(pCollisionRc, tx+1, ty+1, pCollisionList);
		}

	/* Check dynamic objects */
	pObj =  TheGame.curr_level.pObjList;
	while(pObj)
		{
		if (pCollisionList->bCheckObjects[pObj->VTable.GetType(pObj)] == FALSE)
			{
			/* This object has been ignored */
			}
		else if (pObj != pCollisionList->pIgnoreObj)
			{
			if (pObj->VTable.ProcessCollision)
				{
				exProcessCollisionObject(pObj, pCollisionList);
				}
			else	/* no collision override - assuming a bounding box */
				{
				nme.x = pObj->x;
				nme.y = pObj->y;
				nme.w = TheGame.iTileWidth;
				nme.h = TheGame.iTileHeight;
				if (exCCBoundingBox(pCollisionRc, &nme, pObj, pCollisionList))
					{
					}
				}
			}
		pObj = pObj->pNext;
		}

return pCollisionList->iSize;
}


Uint8 *exCUBuildCollisionSurface(SDL_Surface *pCurr)
{
SDL_PixelFormat *fmt;
Uint8 *ptr, *collptr;
Uint8 *pSurfacePtr;
Uint32 colorkey, pixel;
Uint8 iBytesPerPixel;
Uint8 r,g,b;
int x, y;

	if (SDL_LockSurface(pCurr) < 0)
		{
		fprintf(stderr, "Couldn't lock surface: %s\n", SDL_GetError());
		return NULL;
		}

	ptr = (Uint8 *)malloc(pCurr->w * pCurr->h);
	if (ptr == NULL)
		{
		fprintf(stderr, "Couldn't alloc memory for collision surface");
		return NULL;
		}

	/* Don't assume everyone's using bright green for the colour key */
	colorkey = pCurr->format->colorkey;
	iBytesPerPixel = pCurr->format->BytesPerPixel;
	fmt = pCurr->format;	
	pSurfacePtr = (Uint8 *)pCurr->pixels;
	collptr = ptr;
	for(y=0;y<pCurr->h;y++)
		{
		for(x=0;x<pCurr->w;x++)
			{
			switch(iBytesPerPixel)
				{
				/* palletised */
				case	1:	/* 8-bit colour */
						pixel = *(Uint8 *)pSurfacePtr;
						r = fmt->palette->colors[pixel].r;
						g = fmt->palette->colors[pixel].g;
						b = fmt->palette->colors[pixel].b;

						if (SDL_MapRGB(pCurr->format, r,g,b) == colorkey)
							*collptr = 0;
						else
							*collptr = 1;
						break;

				/* packed pixels */
				case	2:	/* 16-bit color */
				case	3:	/* 24-bit color */
				case	4:	/* 32-bit color */
						if (iBytesPerPixel == 2)	pixel = *(Uint16 *)pSurfacePtr;
						if (iBytesPerPixel == 3)	pixel = (*(Uint32 *)pSurfacePtr)&0xffffff;
						if (iBytesPerPixel == 4)	pixel = *(Uint32 *)pSurfacePtr;
	
						r = ((pixel & fmt->Rmask) >> fmt->Rshift) << fmt->Rloss;
						g = ((pixel & fmt->Gmask) >> fmt->Gshift) << fmt->Gloss;
						b = ((pixel & fmt->Bmask) >> fmt->Bshift) << fmt->Bloss;

						if (SDL_MapRGB(pCurr->format, r,g,b) == colorkey)
							*collptr = 0;
						else
							*collptr = 1;
						break;

				default:	/* any other oddities! */
						*collptr = 1;
						break;
				}

			collptr++;
			pSurfacePtr += iBytesPerPixel;
			}
		pSurfacePtr -= pCurr->w * iBytesPerPixel;
		pSurfacePtr += pCurr->pitch;	/* is already in bytes */
		}

	SDL_UnlockSurface(pCurr);

return ptr;
}


void exCUDestroyCollisionSurface(Uint8 *pCurr)
{
	free(pCurr);
}

BOOL 	exCCBoundingBox2Pixel(const SDL_Rect *pRc, const SDL_Rect *pEnemyRc, const Uint8 *pColSurface, int iRegion, const EX_OBJECT *pObj, EX_COLLISION_LIST *pCList)
{
int x, x1, x2;
int y, y1, y2;
const Uint8 *pColData;

	/* Do the easy bounds check first */
	if (!exDoBoxesIntersect(pRc, pEnemyRc))
		return FALSE;

	if (pColSurface == NULL)
		{
		/* Should never happen, unless surface couldn't be locked */
		/* or we ran out of memory */
		exCCBoundingBox(&pCList->rc, pEnemyRc, pObj, pCList);
		return TRUE;
		}

	/* Find the union of the rectangles */
	/* The numbers we compute are relative to the enemy image, so 0,0 */
	/* Means the top left of the enemy graphic */

	if (pEnemyRc->x >= pRc->x)	/* Enemy to right */
		{
		x1 = 0;
		x2 = (pRc->x+pRc->w) - pEnemyRc->x;
		}
	else       				/* Enemy to left */
		{
		x1 = pRc->x - pEnemyRc->x;
		x2 = pEnemyRc->w;
		}

	if (pEnemyRc->y <= pRc->y)	/* Enemy above */
		{
		y1 = pRc->y - pEnemyRc->y;
		y2 = pEnemyRc->h;
		}
	else
		{
		y1 = 0;
		y2 = (pRc->y+pRc->h) - pEnemyRc->y;
		}

	/* Scan the image from y1 (inclusive) to y2 (exclusive) */
	/* If there are any non-transparent pixels, then the box */
	/* collides with data, and we store the collision */
	pColData = pColSurface;
	pColData += (iRegion%TheGame.iNumTileWidth) * TheGame.iTileWidth;
	pColData += (iRegion/TheGame.iNumTileWidth) * TheGame.iTileHeight * 640;
	pColData += x1;
	pColData += y1 * 640;

	for(y=y1;y<y2;y++)
		{
		for(x=x1;x<x2;x++)
			{
			if (*pColData)
				{
				EX_COLLISION col;

				col.pObj = (EX_OBJECT *)pObj;
				col.edge = 0;
				exAddCollision(pCList, &col);
				return TRUE;
				}
			pColData++;
			}
		pColData -= (x2-x1);
		pColData += 640;
		}
			
return FALSE;
}

